/* eslint-disable no-console */

export const _test = {}
export const _version = '2023.10.13'

/**
 * ```ts
 * export type EscapeMap = Map<any, (value?: any) => string>;
 * ```
 */

/**
 * @typedef {Map<any, (value?: any) => string>} EscapeMap
 *
 * The type of arguments of the `escapeValues` and `escape` helpers.
 */

/**
 * ```ts
 * export type HArgs = [
 *   string | Node,
 *   ...(Record<string, any> | null | undefined | Node | string | number | HArgs)[]
 * ];
 * ```
 */

/**
 * @typedef {[
 *   string | Node,
 *   ...(Record<string, any> | null | undefined | Node | string | number | HArgs)[]
 * ]} HArgs
 *
 * The type of arguments of the `h` and `s` helpers.
 */

/**
 * ```ts
 * export type JcssNode = {
 *   [attributeOrSelector: string]: string | number | JcssNode;
 * };
 * ```
 */

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | JcssNode;
 * }} JcssNode
 *
 * The type of arguments of the `jcss` helper.
 */

/**
 * ```ts
 * export type JcssRoot = Record<string, JcssNode>;
 * ```
 */

/**
 * @typedef {Record<string, JcssNode>} JcssRoot
 *
 * The type of arguments of the `jcss` helper.
 */

const _COLORS = ['#e22', '#e73', '#fc3', '#ad4', '#4d9', '#3be', '#45d', '#c3e']

/**
 * ```ts
 * export function chartable(options?: {
 *   bottom?: number;
 *   gapX?: number;
 *   gapY?: number;
 *   headerColumn?: boolean;
 *   headerRow?: boolean;
 *   id?: string;
 *   left?: number;
 *   maxY?: number;
 *   right?: number;
 *   singleScale?: boolean;
 *   table?: HTMLTableElement;
 *   title?: string;
 *   top?: number;
 *   xLabels?: string[];
 *   zLabels?: string[];
 *   zxY?: number[][];
 * }): SVGSVGElement;
 * ```
 */

/**
 * A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
 *
 * Options:
 * - `bottom`: bottom padding (for X axis labels)
 * - `gapX`: X axis spacing
 * - `gapY`: Y axis spacing
 * - `headerColumn`: flag indicating that `table` has a header column (with X axis labels)
 * - `headerRow`: flag indicating that `table` has a header row (with data series labels)
 * - `id`: chart id
 * - `left`: left padding (for data series labels)
 * - `maxY`: number of Y axis lines
 * - `right`: right padding (for data series labels)
 * - `singleScale`: flag to force single scale
 * - `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
 * - `title`: chart title
 * - `top`: top padding (for the title)
 * - `xLabels`: X axis labels
 * - `zLabels`: data series labels
 * - `zxY`: chart data
 *
 * @param {{
 *   bottom?: number;
 *   gapX?: number;
 *   gapY?: number;
 *   headerColumn?: boolean;
 *   headerRow?: boolean;
 *   id?: string;
 *   left?: number;
 *   maxY?: number;
 *   right?: number;
 *   singleScale?: boolean;
 *   table?: HTMLTableElement;
 *   title?: string;
 *   top?: number;
 *   xLabels?: string[];
 *   zLabels?: string[];
 *   zxY?: number[][];
 * }} options
 */
export const chartable = ({
  bottom = 50,
  gapX = 70,
  gapY = 30,
  headerColumn = false,
  headerRow = true,
  id,
  left = 200,
  maxY = 10,
  right = 200,
  singleScale = false,
  table,
  title,
  top = 70,
  xLabels = [],
  zLabels = [],
  zxY = []
} = {}) => {
  if (table != null) {
    const zxYNotPassed = !zxY.length
    const xLabelsNotPassed = !xLabels.length
    const zLabelsNotPassed = !zLabels.length

    table.querySelectorAll('tr').forEach((row, r) => row.querySelectorAll('td,th').forEach((col, c) => {
      const x = r - +headerRow
      const z = c - +headerColumn
      // @ts-expect-error
      const value = col.innerText

      if (x >= 0 && z >= 0 && zxYNotPassed) {
        zxY[z] = zxY[z] ?? []
        zxY[z][x] = parseFloat(value)
      } else if (x >= 0 && z < 0 && xLabelsNotPassed) {
        xLabels[x] = value
      } else if (x < 0 && z >= 0 && zLabelsNotPassed) {
        zLabels[z] = value
      }
    }))
  }

  let bestScales = [Infinity, -Infinity, Infinity, -Infinity, Infinity]

  const ranges = zxY.map(xY => {
    xY = xY.filter(y => !isNaN(y))
    return [Math.min(...xY), Math.max(...xY)]
  })

  if (singleScale) {
    bestScales = ranges.reduce(([totalMin, totalMax], [min, max]) =>
      [Math.min(totalMin, min), Math.max(totalMax, max)], bestScales)
  } else {
    ranges.forEach(([min1, max0]) => ranges.forEach(([min0, max1]) => {
      if (min0 >= min1 && max1 >= max0) {
        let min2 = Infinity
        let max2 = -Infinity

        ranges.forEach(([min, max]) => {
          if (min < min1 || max1 < max) {
            if (min2 > min) {
              min2 = min
            }
            if (max2 < max) {
              max2 = max
            }
          }
        })

        let cost = 0

        ranges.forEach(([min, max]) => {
          if (min < max) {
            cost += 1 - (max - min) / (min1 <= min && max <= max1 ? max1 - min1 : max2 - min2)
          }
        })

        if (bestScales[4] > cost) {
          bestScales = [min1, max1, min2, max2, cost]
        }
      }
    }))
  }

  const [min1, max1, min2, max2] = bestScales
  const maxX = zxY[0].length
  const xi = Array.from({ length: maxX }, (_, x) => x * gapX)
  const yi = Array.from({ length: maxY }, (_, y) => y * gapY)
  const w = gapX * (maxX - 1)
  const h = gapY * (maxY - 1)
  const svgW = left + w + right
  const svgH = top + h + bottom
  const /** @type {HArgs} */ graph = ['g', { transform: `translate(${left} ${top})` },
    ...yi.map(y => ['line', { x1: 0, x2: w, y1: y, y2: y, stroke: '#8888' }]),
    ...xi.map(x => ['line', { x1: x, x2: x, y1: 0, y2: h, stroke: '#8888' }])]
  const /** @type {HArgs} */ lColors = ['g', { class: 'chartable-z-colors', transform: 'translate(-80 0)' }]
  const /** @type {HArgs} */ lLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(-95 0)' }]
  const /** @type {HArgs} */ rColors = ['g', { class: 'chartable-z-colors', transform: 'translate(80 0)' }]
  const /** @type {HArgs} */ rLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(95 0)' }]
  const yLabel = (/** @type {number} */ min, /** @type {number} */ max, /** @type {number} */ y) =>
    (min + (max - min) * (maxY - 1 - y) / (maxY - 1)).toFixed(2).replace(/\.?0+$/, '')

  zxY.forEach((xa, z) => {
    const cls = `chartable-z-${z + 1}`
    const fill = _COLORS[z % _COLORS.length]
    let [min, max] = ranges[z]

    if (min1 <= min && max <= max1) {
      min = min1
      max = max1
      lColors.push(['circle', { class: cls, cx: 0, cy: gapY * (lColors.length - 2), r: 5, fill }])
      lLabels.push(['text',
        { x: 0, y: gapY * (lLabels.length - 2), 'text-anchor': 'end', 'alignment-baseline': 'middle' },
        zLabels[z]])
    } else {
      min = min2
      max = max2
      rColors.push(['circle', { class: cls, cx: 0, cy: gapY * (rColors.length - 2), r: 5, fill }])
      rLabels.push(['text',
        { x: 0, y: gapY * (rLabels.length - 2), 'text-anchor': 'start', 'alignment-baseline': 'middle' },
        zLabels[z]])
    }

    const y = xa.map(x => isNaN(x) ? x : h * (max - x) / (max - min))

    xi.forEach((x, i) => {
      if (!isNaN(y[i])) {
        graph.push(['circle', { class: cls, cx: x, cy: y[i], r: 5, fill }])

        if (!isNaN(y[i - 1])) {
          graph.push(['line', { class: cls, x1: x, x2: xi[i - 1], y1: y[i], y2: y[i - 1], stroke: fill }])
        }
      }
    })
  })

  return s('svg', {
    viewBox: `0 0 ${svgW} ${svgH}`,
    width: `${svgW}px`,
    height: `${svgH}px`,
    class: 'chartable'
  }, id != null ? { id } : null, title != null
    ? ['text',
        { class: 'chartable-title', x: svgW / 2, y: top / 2, 'text-anchor': 'middle', 'alignment-baseline': 'middle' },
        title]
    : null,
  graph,
  min1 < Infinity
    ? ['g', { class: 'chartable-left', transform: `translate(${left} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(-15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'end', 'alignment-baseline': 'middle' },
            yLabel(min1, max1, i)])],
        lColors, lLabels]
    : null,
  min2 < Infinity
    ? ['g', { class: 'chartable-right', transform: `translate(${left + w} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'start', 'alignment-baseline': 'middle' },
            yLabel(min2, max2, i)])],
        rColors, rLabels]
    : null,
  xLabels.length > 0
    ? ['g', { transform: `translate(${left} ${top + h})` },
        ['g', { class: 'chartable-x-labels', transform: 'translate(0 15)' },
          ...xi.map((x, i) => ['text',
            { x, y: 0, 'text-anchor': 'middle', 'alignment-baseline': 'hanging' },
            xLabels[i]])]]
    : null
  )
}

/**
 * ```ts
 * export function eq(x: any, y: any): boolean;
 * ```
 */

/**
 * A helper that checks equality of the given arguments.
 */
export const eq = /** @returns {boolean} */ (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const xConstructor = x?.constructor

  if (xConstructor === y?.constructor) {
    if (xConstructor === Number) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (xConstructor === Date) {
      return +x === +y
    }

    if (xConstructor === String || xConstructor === RegExp) {
      return '' + x === '' + y
    }

    if (xConstructor === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => eq(item, y[index]))
    }

    if (xConstructor === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => eq(x[key], y[key]))
    }

    if (xConstructor === Set || xConstructor === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => eq(xv, yv)))
    }
  }

  return false
}

_test.eq = () => {
  console.assert(eq(true, true))
  console.assert(eq(NaN, NaN))
  console.assert(!eq(null, undefined))
  console.assert(eq(42, 42))
  // eslint-disable-next-line no-new-wrappers
  console.assert(eq(42, new Number(42)))
  console.assert(eq(42, Number(42)))
  // eslint-disable-next-line no-new-wrappers
  console.assert(eq(new Number(42), Number(42)))
  console.assert(!eq(42, '42'))
  console.assert(eq('42', '42'))
  // eslint-disable-next-line no-new-wrappers
  console.assert(eq('42', new String('42')))
  console.assert(eq('42', String('42')))
  // eslint-disable-next-line no-new-wrappers
  console.assert(eq(String('42'), new String('42')))
  console.assert(eq(/42/, /42/))
  console.assert(!eq(/42/, /42/g))
  console.assert(eq(new Date(42), new Date(42)))
  console.assert(!eq(new Date(), new Date(42)))
  console.assert(eq({ j: '42', c: 42 }, { c: 42, j: '42' }))
  console.assert(eq([42, '42'], [42, '42']))
  console.assert(eq(new Set(['42', 42]), new Set([42, '42'])))
  console.assert(!eq(new Set(['42', 42]), new Set([42])))
  console.assert(!eq(new Set([42, undefined]), new Set([42])))
  console.assert(eq(
    new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]),
    new Map([[{ c: 42 }, { C: '42' }], [{ j: 42 }, { J: '42' }]])
  ))
  console.assert(!eq(
    new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]),
    new Map([[{ j: '42' }, { J: 42 }], [{ c: '42' }, { C: 42 }]])
  ))
}

/**
 * ```ts
 * export function escape(escapeMap: EscapeMap, template: TemplateStringsArray, ...values: any[]): string;
 * ```
 */

/**
 * A generic helper for escaping `values` by given `escapeMap` (in *TemplateStrings* flavor).
 */
export const escape = (
  /** @type {EscapeMap}            */ escapeMap,
  /** @type {TemplateStringsArray} */ template,
  /** @type {any[]}                */ ...values
) => String.raw(template, ...escapeValues(escapeMap, values))

_test.escape = () => {
  // @ts-expect-error
  const /** @type {EscapeMap} */ escapeMap = new Map([
    [undefined, () => 'NULL'],
    [Array, (/** @type {any[]} */ values) => escapeValues(escapeMap, values).join(', ')],
    [Boolean, (/** @type {boolean} */ value) => `b'${+value}'`],
    [Date, (/** @type {Date} */ value) => `'${value.toISOString().replace(/^(.+)T(.+)\..*$/, '$1 $2')}'`],
    [Number, (/** @type {number} */ value) => `${value}`],
    [String, (/** @type {string} */ value) => `'${value.replace(/'/g, "''")}'`]
  ])

  // @ts-expect-error
  const sql = escape.bind(null, escapeMap)

  const actual = sql`
    SELECT *
    FROM table_name
    WHERE column_name IN (${[true, null, undefined, 42, '42', "4'2", /42/, new Date(323325e6)]})`

  const expected = `
    SELECT *
    FROM table_name
    WHERE column_name IN (b'1', NULL, NULL, 42, '42', '4''2', NULL, '1980-03-31 04:30:00')`

  console.assert(actual === expected)
}

/**
 * ```ts
 * export function escapeValues(escapeMap: EscapeMap, values: any[]): string[];
 * ```
 */

/**
 * A generic helper for escaping `values` by given `escapeMap`.
 */
export const escapeValues = (
  /** @type {EscapeMap} */ escapeMap,
  /** @type {any[]}     */ values
) => values.map(value => (escapeMap.get(value?.constructor) ?? escapeMap.get(undefined))?.(value) ?? '')

const _TAGS_TO_SKIP = { IFRAME: 1, NOSCRIPT: 1, PRE: 1, SCRIPT: 1, STYLE: 1, TEXTAREA: 1 }

/**
 * ```ts
 * export function fixTypography(node: Node): void;
 * ```
 */

/**
 * A helper that implements typographic corrections specific to Polish typography.
 */
export const fixTypography = (/** @type {Node} */ node) => {
  const /** @type {Node[]} */ queue = [node]

  while (queue.length > 0) {
    const node0 = queue.shift()

    if (node0 instanceof Element) {
      for (let i = 0; i < node0.childNodes.length; ++i) {
        const childNode = node0.childNodes[i]

        if (childNode instanceof Text) {
          queue.push(childNode)
        } else if (childNode instanceof Element && !has(childNode.tagName, _TAGS_TO_SKIP)) {
          queue.push(childNode)
        }
      }
    } else if (node0 instanceof Text && node0.nodeValue != null && node0.nodeValue.trim() !== '') {
      let /** @type {Node} */ previousNode = node0

      node0.nodeValue.split(/(\s|\(|„)([aiouwz—]\s)/gi).forEach((chunk, i) => {
        i %= 3

        const currentNode = i === 2
          ? h('span', { style: 'white-space:nowrap' }, chunk)
          : i === 1
            ? document.createTextNode(chunk)
            : document.createTextNode(chunk.replace(/(\/(?=[^/\s])|\.(?=[^\s]))/g, '$1\u200B'))

        if (node0.parentNode != null) {
          node0.parentNode.insertBefore(currentNode, previousNode.nextSibling)
        }

        previousNode = currentNode
      })

      node0.parentNode?.removeChild(node0)
    }
  }
}

_test.fixTypography = () => {
  const p = h('p', 'Pchnąć w tę łódź jeża lub ośm skrzyń fig (zob. https://pl.wikipedia.org/wiki/Pangram).')

  fixTypography(p)

  console.assert(p.innerHTML ===
    'Pchnąć <span style="white-space:nowrap">w </span>tę łódź jeża lub ośm skrzyń fig ' +
    '(zob. https://\u200Bpl.\u200Bwikipedia.\u200Borg/\u200Bwiki/\u200BPangram).')
}

const /** @type {Record<string, string>} */ _NS = {
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * @type {{
 *   (namespaceURI?: null | undefined): {
 *     <T extends keyof HTMLElementTagNameMap>(tag:  T, ...args: HArgs[1][]): HTMLElementTagNameMap[T];
 *     <N extends Node>                       (node: N, ...args: HArgs[1][]): N;
 *     (...args: HArgs): Node;
 *   };
 *   (namespaceURI: 'http://www.w3.org/2000/svg'): {
 *     <T extends keyof SVGElementTagNameMap> (tag:  T, ...args: HArgs[1][]): SVGElementTagNameMap[T];
 *     <N extends Node>                       (node: N, ...args: HArgs[1][]): N;
 *     (...args: HArgs): Node;
 *   };
 * }}
 */
const _h = (/** @type {string?=} */ namespaceURI) => {
  const createElement = namespaceURI == null
    ? (/** @type {string } */ tag) => document.createElement(tag)
    : (/** @type {string } */ tag) => document.createElementNS(namespaceURI, tag)

  const h = (/** @type {HArgs[0]} */ tagOrNode, /** @type {HArgs[1][]} */ ...args) => {
    const node = is(String, tagOrNode) ? createElement(tagOrNode) : tagOrNode

    for (const arg of args) {
      let child = null

      if (arg instanceof Node) {
        child = arg
      } else if (is(String, arg) || is(Number, arg)) {
        // @ts-expect-error
        child = new Text(arg)
      } else if (is(Array, arg)) {
        // @ts-expect-error
        child = h(...arg)
      } else if (arg != null) {
        for (const name in arg) {
          const value = arg[name]

          if (name[0] === '$') {
            const name1 = name.slice(1)

            if (is(Object, value)) {
              // @ts-expect-error
              node[name1] = node[name1] ?? {}
              // @ts-expect-error
              Object.assign(node[name1], value)
            } else {
              // @ts-expect-error
              node[name1] = value
            }
          } else if (node instanceof Element) {
            const indexOfColon = name.indexOf(':')

            if (indexOfColon >= 0) {
              const /** @type {string=} */ ns = _NS[name.slice(0, indexOfColon)]

              if (ns != null) {
                const basename = name.slice(indexOfColon + 1)

                if (value === true) {
                  node.setAttributeNS(ns, basename, '')
                } else if (value === false) {
                  node.removeAttributeNS(ns, basename)
                } else {
                  node.setAttributeNS(ns, basename, is(String, value) ? value : '' + value)
                }
              }
            } else {
              if (value === true) {
                node.setAttribute(name, '')
              } else if (value === false) {
                node.removeAttribute(name)
              } else {
                node.setAttribute(name, is(String, value) ? value : '' + value)
              }
            }
          }
        }
      }

      if (child != null) {
        node.appendChild(child)
      }
    }

    return node
  }

  return h
}

/**
 * ```ts
 * export function h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args: HArgs[1][]): HTMLElementTagNameMap[T];
 * export function h<N extends Node>(node: N, ...args: HArgs[1][]): N;
 * export function h(...args: HArgs): Node;
 * ```
 */

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `s`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttribute`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `h` and the results are appended to the element being created or modified.
 */
export const h = _h()

_test.h = () => {
  const b = h('b')

  console.assert(b.outerHTML === '<b></b>')

  const i = h('i', 'text')

  h(b, i)

  console.assert(i.outerHTML === '<i>text</i>')
  console.assert(b.outerHTML === '<b><i>text</i></b>')

  h(i, { $className: 'some class' })

  console.assert(i.outerHTML === '<i class="some class">text</i>')
  console.assert(b.outerHTML === '<b><i class="some class">text</i></b>')
}

_test['h: innerText vs items'] = () => {
  console.assert(h('span', 'text').outerHTML === '<span>text</span>')
  console.assert(h('span', { $innerText: 'text' }).outerHTML === '<span>text</span>')
}

_test['h: style'] = () => {
  console.assert(h('div', { style: 'margin:0;padding:0' }).outerHTML ===
    '<div style="margin:0;padding:0"></div>')
  console.assert(h('div', { $style: 'margin:0;padding:0' }).outerHTML ===
    '<div style="margin: 0px; padding: 0px;"></div>')
  console.assert(h('div', { $style: { margin: 0, padding: 0 } }).outerHTML ===
    '<div style="margin: 0px; padding: 0px;"></div>')
}

_test['h: attributes vs properties'] = () => {
  const input1 = h('input', { value: 42 })
  const input2 = h('input', { $value: '42' })

  console.assert(input1.value === '42')
  console.assert(input2.value === '42')

  console.assert(input1.outerHTML === '<input value="42">')
  console.assert(input2.outerHTML === '<input>')

  const checkbox1 = h('input', { type: 'checkbox', checked: true })
  const checkbox2 = h('input', { type: 'checkbox', $checked: true })

  console.assert(checkbox1.checked === true)
  console.assert(checkbox2.checked === true)

  console.assert(checkbox1.outerHTML === '<input type="checkbox" checked="">')
  console.assert(checkbox2.outerHTML === '<input type="checkbox">')
}

_test['h: nested properties'] = () => {
  const div = h('div')

  // @ts-expect-error
  console.assert(div.key === undefined)

  h(div, { $key: { one: 1 } })

  // @ts-expect-error
  console.assert(eq(div.key, { one: 1 }))

  h(div, { $key: { two: 2 } })

  // @ts-expect-error
  console.assert(eq(div.key, { one: 1, two: 2 }))
}

/**
 * ```ts
 * export function has(key: any, ref: any): boolean;
 * ```
 */

/**
 * A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export const has = (/** @type {any} */ key, /** @type {any} */ ref) =>
  (is(String, key) || is(Number, key) || is(Symbol, key)) && Object.hasOwnProperty.call(ref ?? Object, key)

_test.has = () => {
  const obj = { key: 'K', null: 'N' }

  console.assert('key' in obj)
  console.assert(has('key', obj))

  console.assert('null' in obj)
  console.assert(has('null', obj))

  // @ts-expect-error
  console.assert(null in obj)
  console.assert(!has(null, obj))

  console.assert('toString' in obj)
  console.assert(!has('toString', obj))
}

_test['has: null'] = () => {
  let typeError

  try {
    // @ts-expect-error
    console.assert('key' in null)
  } catch (error) {
    typeError = error
  }

  console.assert(typeError instanceof TypeError) // Cannot use 'in' operator to search for 'key' in null
  console.assert(!has('key', null))
}

/**
 * ```ts
 * export function is(type: BigIntConstructor, arg: any): arg is bigint;
 * export function is(type: BooleanConstructor, arg: any): arg is boolean;
 * export function is(type: NumberConstructor, arg: any): arg is number;
 * export function is(type: StringConstructor, arg: any): arg is string;
 * export function is(type: SymbolConstructor, arg: any): arg is symbol;
 * export function is(type: undefined, arg: any): arg is null | undefined;
 * export function is<T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * ```
 */

/**
 * A helper that checks if the given argument is of a certain type.
 *
 * @template {abstract new (...args: any[]) => any} T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @returns {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

_test.is = () => {
  console.assert(is(Number, 42))
  console.assert(is(Number, Number(42)))
  // eslint-disable-next-line no-new-wrappers
  console.assert(is(Number, new Number(42)))
  console.assert(is(Number, NaN))
  console.assert(is(String, '42'))
  console.assert(is(String, String('42')))
  // eslint-disable-next-line no-new-wrappers
  console.assert(is(String, new String('42')))
  console.assert(is(Symbol, Symbol('42')))
  console.assert(is(Symbol, Object(Symbol('42'))))
  console.assert(is(undefined, undefined))
  console.assert(is(undefined, null))
  console.assert(is(Object, {}))
  console.assert(is(Array, []))
  console.assert(is(RegExp, /42/))
  console.assert(is(Date, new Date(42)))
  console.assert(is(Set, new Set(['42', 42])))
  console.assert(is(Map, new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]])))
}

_test['is vs ‘toString.call’'] = () => {
  class FooBar { }

  console.assert(is(FooBar, new FooBar()))

  const fakeFooBar = { [Symbol.toStringTag]: 'FooBar' }

  console.assert(({}).toString.call(new FooBar()) === '[object Object]')
  console.assert(({}).toString.call(fakeFooBar) === '[object FooBar]')
}

const _jcss = (
  /** @type {JcssNode} */ node,
  /** @type {string}   */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[JcssNode | string[], string][]} */ queue = [[node, prefix]]

  while (queue.length) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if (is(Array, style2)) {
      result.push(prefix2, prefix2 !== '' ? '{' : '', style2.join(';'), prefix2 !== '' ? '}' : '')
    } else {
      const /** @type {[JcssNode | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style2) {
        const value = style2[key]

        if (is(String, value) || is(Number, value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * ```ts
 * export function jcss(root: JcssRoot, splitter?: string): string;
 * ```
 */

/**
 * A simple CSS-in-JS helper.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export const jcss = (/** @type {JcssRoot} */ root, splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in root) {
    const value = root[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _jcss(value, '', result, split)
      result.push('}')
    } else {
      _jcss(value, split(key), result, split)
    }
  }

  return result.join('')
}

_test['jcss: #1'] = () => {
  const actual = jcss({
    a: {
      color: 'red',
      margin: 1,
      '.c': { margin: 2, padding: 2 },
      padding: 1
    }
  })

  const expected = `
    a{
      color:red;
      margin:1
    }
    a.c{
      margin:2;
      padding:2
    }
    a{
      padding:1
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

_test['jcss: #2'] = () => {
  const actual = jcss({
    a: {
      '.b': {
        color: 'red',
        margin: 1,
        '.c': { margin: 2, padding: 2 },
        padding: 1
      }
    }
  })

  const expected = `
    a.b{
      color:red;
      margin:1
    }
    a.b.c{
      margin:2;
      padding:2
    }
    a.b{
      padding:1
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

_test['jcss: #3'] = () => {
  const actual = jcss({
    '@font-face$$1': {
      fontFamily: 'Jackens',
      src$$1: 'url(otf/jackens.otf)',
      src$$2: "url(otf/jackens.otf) format('opentype')," +
        "url(svg/jackens.svg) format('svg')",
      fontWeight: 'normal',
      fontStyle: 'normal'
    },
    '@font-face$$2': {
      fontFamily: 'C64',
      src: 'url(fonts/C64_Pro_Mono-STYLE.woff)'
    },
    '@keyframes spin': {
      '0%': { transform: 'rotate(0deg)' },
      '100%': { transform: 'rotate(360deg)' }
    },
    div: {
      border: 'solid red 1px',
      '.c1': { 'background-color': '#000' },
      ' .c1': { backgroundColor: 'black' },
      '.c2': { backgroundColor: 'rgb(0,0,0)' }
    },
    '@media(min-width:200px)': {
      div: { margin: 0, padding: 0 },
      span: { color: '#000' }
    }
  })

  const expected = `
    @font-face{
      font-family:Jackens;
      src:url(otf/jackens.otf);
      src:url(otf/jackens.otf) format('opentype'),url(svg/jackens.svg) format('svg');
      font-weight:normal;
      font-style:normal
    }
    @font-face{
      font-family:C64;
      src:url(fonts/C64_Pro_Mono-STYLE.woff)
    }
    @keyframes spin{
      0%{
        transform:rotate(0deg)
      }
      100%{
        transform:rotate(360deg)
      }
    }
    div{
      border:solid red 1px
    }
    div.c1{
      background-color:#000
    }
    div .c1{
      background-color:black
    }
    div.c2{
      background-color:rgb(0,0,0)
    }
    @media(min-width:200px){
      div{
        margin:0;
        padding:0
      }
      span{
        color:#000
      }
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

_test['jcss: #4'] = () => {
  const actual = jcss({
    a: {
      '.b,.c': {
        margin: 1,
        '.d': {
          margin: 2
        }
      }
    }
  })

  const expected = `
    a.b,a.c{
      margin:1
    }
    a.b.d,a.c.d{
      margin:2
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

_test['jcss: #5'] = () => {
  const actual = jcss({
    '.b,.c': {
      margin: 1,
      '.d': {
        margin: 2
      }
    }
  })

  const expected = `
    .b,.c{
      margin:1
    }
    .b.d,.c.d{
      margin:2
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

_test['jcss: #6'] = () => {
  const actual = jcss({
    '.a,.b': {
      margin: 1,
      '.c,.d': {
        margin: 2
      }
    }
  })

  const expected = `
    .a,.b{
      margin:1
    }
    .a.c,.a.d,.b.c,.b.d{
      margin:2
    }`.replace(/\n\s*/g, '')

  console.assert(actual === expected)
}

/**
 * ```ts
 * export function jsOnParse(handlers: Record<string, Function>, text: string): any;
 * ```
 */

/**
 * `JSON.parse` with “JavaScript turned on”.
 *
 * Objects having *exactly* one property which is present in the `handlers` map, i.e. objects of the form:
 *
 * ```js
 * { "«handlerName»": [«params»] }
 * ```
 *
 * are replaced by the result of call
 *
 * ```js
 * handlers['«handlerName»'](...«params»)
 * ```
 */
export const jsOnParse = (
  /** @type {Record<string, Function>} */ handlers,
  /** @type {string}                   */ text
) => JSON.parse(text, (key, value) => {
  if (is(Object, value)) {
    let isSecondKey = false

    for (key in value) {
      if (isSecondKey) {
        return value
      }
      isSecondKey = true
    }

    if (has(key, handlers) && is(Array, value[key])) {
      return handlers[key](...value[key])
    }
  }

  return value
})

_test.jsOnParse = () => {
  const handlers = {
    $hello: (/** @type {string} */ name) => `Hello ${name}!`,
    $foo: () => 'bar'
  }
  const actual = jsOnParse(handlers, `
  [
    {
      "$hello": ["World"]
    },
    {
      "nested": {
        "$hello": ["nested World"]
      },
      "one": 1,
      "two": 2
    },
    {
      "$foo": []
    },
    {
      "$foo": ["The parent object does not have exactly one property!"],
      "one": 1,
      "two": 2
    }
  ]`)
  const expected = [
    'Hello World!',
    {
      nested: 'Hello nested World!',
      one: 1,
      two: 2
    },
    'bar',
    {
      $foo: ['The parent object does not have exactly one property!'],
      one: 1,
      two: 2
    }]

  console.assert(eq(actual, expected))
}

const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => {
  // @ts-expect-error
  const v = locales?.[language]?.[version]?.[text]

  if (is(String, v)) {
    return v
  }

  const t = locales?.[language]?.[text]

  return is(String, t) ? t : text
}

/**
 * ```ts
 * export function locale(
 *   locales: Record<string, Record<string, string | Record<string, string>>>,
 *   defaultLanguage: string,
 *   languages?: readonly string[]
 * ): (text?: string, version?: string) => string;
 * ```
 */

/**
 * Language translations helper.
 */
export const locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  languages = navigator.languages
) => {
  for (const language of languages) {
    if (has(language, locales)) {
      // @ts-expect-error
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error
  return _locale.bind(0, locales, defaultLanguage)
}

_test.locale = () => {
  const locales = {
    pl: {
      Password: 'Hasło',
      button: { Login: 'Zaloguj' }
    }
  }
  const _ = locale(locales, 'pl', [])

  console.assert(_('Login') === 'Login')
  console.assert(_('Password') === 'Hasło')

  console.assert(_('Undefined text') === 'Undefined text')

  console.assert(_('Login', 'button') === 'Zaloguj')

  console.assert(_('Password', 'undefined_version') === 'Hasło')
  console.assert(_('Undefined text', 'undefined_version') === 'Undefined text')

  console.assert(_('toString') === 'toString')
  console.assert(_('toString', 'undefined_version') === 'toString')
}

/**
 * ```ts
 * export function nanolight(
 *   pattern: RegExp,
 *   highlighters: ((chunk: string, index: number) => HArgs[1])[],
 *   code: string
 * ): HArgs[1][];
 * ```
 */

/**
 * A generic helper for syntax highlighting (see also `nanolightJs`).
 */
export const nanolight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => HArgs[1])[]} */ highlighters,
  /** @type {string} */ code
) => {
  const /** @type {HArgs[1][]} */ result = []

  code.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

/**
 * ```ts
 * export function nanolightJs(codeJs: string): HArgs[1][];
 * ```
 */

/**
 * A helper for highlighting JavaScript.
 */
// @ts-expect-error
export const nanolightJs = nanolight.bind(0,
  /('.*?'|".*?"|`[\s\S]*?`)|(\/\/.*?\n|\/\*[\s\S]*?\*\/)|(break|case|catch|const|continue|debugger|default|delete|do|else|eval|export\s+type|export|extends|false|finally|for|from|function|goto|if|import|in|instanceof|is|keyof|let|NaN|new|null|package|return|super|switch|this|throw|true|try|typeof|undefined|var|void|while|with|yield)(?!\w)|([<>=.?:&|!~*/%+-])|(0x[\dabcdef]+|0o[01234567]+|0b[01]+|\d+(?:\.[\d_]+)?(?:e[+-]?[\d_]+)?)|([$\w]+)(?=\()|([$\wąćęłńóśżźĄĆĘŁŃÓŚŻŹ]+)/,
  [
    chunk => chunk, // no match
    chunk => ['u', chunk], // string literals
    chunk => ['em', chunk], // comments
    chunk => ['b', chunk], // keywords
    chunk => ['b', chunk], // operators
    chunk => ['u', chunk], // number literals
    chunk => ['u', chunk], // function calls
    chunk => ['i', chunk] // literals
  ]
)

_test.nanolightJs = () => {
  const codeJs = 'const answerToLifeTheUniverseAndEverything = 42'

  console.assert(h('pre', ['code', ...nanolightJs(codeJs)]).outerHTML ===
    '<pre><code><b>const</b> <i>answerToLifeTheUniverseAndEverything</i> <b>=</b> <u>42</u></code></pre>')
}

/**
 * ```ts
 * export function plUral(singular: string, plural2: string, plural5: string, value: number): string;
 * ```
 */

/**
 * A helper for choosing the correct singular and plural.
 */
export const plUral = (
  /** @type {string} */ singular,
  /** @type {string} */ plural2,
  /** @type {string} */ plural5,
  /** @type {number} */ value
) => {
  const absValue = Math.abs(value)
  const absValueMod10 = absValue % 10

  return value === 1
    ? singular
    : absValue === 12 || absValue === 13 || absValue === 14
      ? plural5
      : absValueMod10 === 2 || absValueMod10 === 3 || absValueMod10 === 4
        ? plural2
        : plural5
}

_test.plUral = () => {
  // @ts-expect-error
  const auto = plUral.bind(null, 'auto', 'auta', 'aut')

  console.assert(auto(0) === 'aut')
  console.assert(auto(1) === 'auto')
  console.assert(auto(17) === 'aut')
  console.assert(auto(42) === 'auta')

  // @ts-expect-error
  const car = plUral.bind(null, 'car', 'cars', 'cars')

  console.assert(car(0) === 'cars')
  console.assert(car(1) === 'car')
  console.assert(car(17) === 'cars')
  console.assert(car(42) === 'cars')
}

/**
 * ```ts
 * export function pro(ref: any): any;
 * ```
 */

/**
 * A helper that protect calls to nested properties by `Proxy` which initializes non-existent values with an empty object.
 */
// @ts-expect-error
export const pro = (/** @type {any} */ ref) => new Proxy(ref, {
  get (target, key) {
    return pro(target[key] = target[key] ?? {})
  }
})

_test.pro = () => {
  const ref = {}

  pro(ref).one.two[3][4] = 1234

  console.assert(eq(ref, { one: { two: { 3: { 4: 1234 } } } }))

  pro(ref).one.two.tree = 123

  console.assert(eq(ref, { one: { two: { 3: { 4: 1234 }, tree: 123 } } }))

  pro(ref).one.two = undefined

  console.assert(eq(ref, { one: { two: undefined } }))

  delete pro(ref).one.two

  console.assert(eq(ref, { one: {} }))

  pro(ref).one.two.three.four

  console.assert(eq(ref, { one: { two: { three: { four: {} } } } }))

  pro(ref).one.two.three.four = 1234

  console.assert(eq(ref, { one: { two: { three: { four: 1234 } } } }))
}

/**
 * ```ts
 * export function refsInfo(...refs: any[]): [string, string, string[]][];
 * ```
 */

/**
 * A helper that provides information about the given `refs`.
 *
 * It returns an array of triples: `[«name», «prototype-name», «array-of-own-property-names»]`.
 */
export const refsInfo = (/** @type {any[]} */ ...refs) => {
  const /** @type {Set<Function>} */ fns = new Set()

  refs.forEach(ref => {
    while (is(Function, ref) && !fns.has(ref) && ref.toString?.()?.includes?.('[native code]')) {
      fns.add(ref)
      ref = Object.getPrototypeOf(ref)
    }
  })

  return Array.from(fns.values()).map(/** @returns {[string, string, string[]]} */ fn => [
    fn.name,
    Object.getPrototypeOf(fn)?.name ?? '',
    Object.getOwnPropertyNames(fn.prototype ?? Object.create(null)).sort()
  ]).sort((a, b) => -(a[0] < b[0]))
}

_test.refsInfo = () => {
  const info = refsInfo(Array, Function)

  console.assert(info.find(([name]) => name === 'Array')?.[2]?.includes?.('length'))
  console.assert(info.find(([name]) => name === 'Function')?.[2]?.includes?.('length'))
}

_test['refsInfo: browserFingerprint'] = () => {
  const browserFingerprint = () => {
    // @ts-expect-error
    const refs = Object.getOwnPropertyNames(window).map(name => window[name])
    const info = refsInfo(...refs)
    const json = JSON.stringify(info)
    const hash = Array(32).fill(0)
    let j = 0

    for (let i = 0; i < json.length; i++) {
      let charCode = json.charCodeAt(i)

      while (charCode > 0) {
        hash[j] = hash[j] ^ (charCode & 15)
        charCode >>= 4
        j = (j + 1) & 31
      }
    }

    return hash.map(x => x.toString(16)).join('')
  }

  console.log(browserFingerprint())
}

/**
 * ```ts
 * export function s<T extends keyof SVGElementTagNameMap>(tag: T, ...args: HArgs[1][]): SVGElementTagNameMap[T];
 * export function s<N extends Node>(node: N, ...args: HArgs[1][]): N;
 * export function s(...args: HArgs): Node;
 * ```
 */

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `h`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttributeNS`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttributeNS`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `s` and the results are appended to the element being created or modified.
 */
export const s = _h('http://www.w3.org/2000/svg')

const _ZEROS = '0'.repeat(16)
let _counter = 0

/**
 * ```ts
 * export function uuid1(options?: {
 *   date?: Date;
 *   node?: string;
 * }): string;
 * ```
 */

/**
 * A helper that generates a UUID v1 identifier (with a creation timestamp).
 *
 * - The optional `node` parameter should have the format `/^[0123456789abcdef]+$/`.
 *   Its value will be trimmed to last 12 characters and left padded with zeros.
 */
export const uuid1 = ({
  date = new Date(),
  node = Math.random().toString(16).slice(2)
} = {}) => {
  const time = _ZEROS + (10_000 * (+date + 12_219_292_800_000)).toString(16)

  _counter = (_counter + 1) & 0x3fff

  return time.slice(-8).concat(
    '-',
    time.slice(-12, -8),
    // @ts-expect-error
    -1,
    time.slice(-15, -12),
    '-',
    (8 | (_counter >> 12)).toString(16),
    (_ZEROS + (_counter & 0xfff).toString(16)).slice(-3),
    '-',
    (_ZEROS + node).slice(-12))
}

_test.uuid1 = () => {
  for (let counter = 1; counter <= 0x5678; ++counter) {
    const uuid = uuid1()

    counter === 0x0001 && console.assert(uuid.split('-')[3] === '8001')
    counter === 0x0fff && console.assert(uuid.split('-')[3] === '8fff')
    counter === 0x1000 && console.assert(uuid.split('-')[3] === '9000')
    counter === 0x2345 && console.assert(uuid.split('-')[3] === 'a345')
    counter === 0x3456 && console.assert(uuid.split('-')[3] === 'b456')
    counter === 0x4000 && console.assert(uuid.split('-')[3] === '8000')
    counter === 0x4567 && console.assert(uuid.split('-')[3] === '8567')
  }
}

_test['uuid1: node'] = () => {
  console.assert(uuid1({ node: '000123456789abc' }).split('-')[4] === '123456789abc')
  console.assert(uuid1({ node: '123456789' }).split('-')[4] === '000123456789')
}

_test['uuid1: date'] = () => {
  console.assert(uuid1({ date: new Date(323325e6) }).startsWith('c1399400-9a71-11bd'))
}
